#
# inetc.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later..
#

require 'net/smtp'
require 'net/pop'

require 'amstd/bug'
require 'amstd/dispatch'


module EST


module EstNet

  class << self
    def append_feature( mod )
      mod.protocol_param :socket_type, '::EST::INETsocket'
    end
  end

end


class INETclient < Strategy

  class << self
    def session_type( val = nil )
      if val then
        @session_type = val
      else
        @session_type
      end
    end
  end


  def initialize( rc )
    super
    @session = type.session_type.new( @rc['address'], @rc['port'].to_i )
  end

  def address
    @session.address
  end

  def port
    @session.port
  end

  def executing?
    @session.active?
  end


  private

  def do_exec( *args )
    if_logging_in
    @session.start( *args )
    if_logged_in
    if_logging_out
    @session.finish
    if_logged_out
  end

  def if_logging_in
  end

  def if_logged_in
  end

  def if_logging_out
  end

  def if_logged_out
  end

end


class INETsocket < Net::Socket

  include Dispatchable

  def initialize( addr, port, pipe = nil )
    super( addr, port, pipe )

    @pipe << "TCP connection to #{addr} opened" if @pipe
  end

  event_handler :on_writing
  event_handler :on_write
  event_handler :on_writeall

  event_handler :on_reading
  event_handler :on_read
  event_handler :on_readall


  private

  def do_write_beg
    super
    dispatch :on_writing
  end

  def do_write_do( arg )
    ret = super
    dispatch :on_write, ret

    ret
  end

  def do_write_fin
    ret = super
    dispatch :on_writeall, ret

    ret
  end


  public

  def read( len, &blk )
    do_read(blk){ super len }
  end

  def readuntil( targ, &blk )
    do_read(blk){ super targ }
  end
      
  def read_pendstr( ret = '', &blk )
    do_read(blk){ super ret }
  end

  def read_pendlist
    do_read(nil){ super }
  end


  private

  def do_read( block )
    push_event( :on_read, block ) if block
    dispatch :on_reading
    ret = yield
    dispatch :on_readall
    pop_event( :on_read ) if block

    ret
  end

  def fill_rbuf
    tmp = @socket.sysread( READ_BLOCK )
    @buffer << tmp
    dispatch :on_read, tmp.size
  end

end



class SMTP < Net::SMTP
  include EstNet
end

class SMTPclient < INETclient

  RC_DEFAULT = {
    'address'      => 'localhost',
    'port'         => '25',
    'hello-domain' => ENV['HOSTNAME'],
    # 'move-to'    => 'draft',
    '.end.'      => nil
  }


  session_type SMTP

  def exec( *indexes )
    if CORE.mailbox.draft.size == 0 then
      if_no_mail
      return
    end

    dra = CORE.mailbox.draft
    if indexes.empty? then
      @send_mails = a = []
      dra.each_with_index {|po,i| a.push [i,po] }
    else
      @send_mails = indexes.collect {|i| [i, dra[i]] }
    end

    do_exec @rc['hello-domain']
  end


  private


  def if_no_mail
    CORE.notice 'no mail to send. cancel'
  end

  def if_logged_in
    CORE.notice "connected to #{address}"
    @n = 0

    draft = CORE.mailbox.draft
    @send_mails.each do |arr|
      begin
        mail = Mail.new( arr[1], true )
        from = from_domain( mail )
        to   = to_addrs( mail )
        src  = ready_to_send( mail )
        @session.sendmail( src, from, to )
        if_sent_mail( mail, arr[0] )
      rescue ArgumentError, Net::ProtocolError
        if_send_fail( mail, arr[0] )
      end
    end
  end

  def if_sent_mail( mail, idx )
    @n += 1
    CORE.notice "sent mail \##{@n}"
  end

  def if_send_fail( mail, idx )
    CORE.notice "send fail: \##{idx}"
  end

  def if_logged_out
    mvto = @rc['move-to']
    return unless mvto

    dra = CORE.mailbox.draft
    while dra[0] do
      CORE.mvmail dra, mvto, 0
    end

    CORE.notice "sent total #{@n} mails"
  end


  ###


  def from_domain( mail )
    unless faddr = mail.from_addrs(nil) and faddr = faddr[0] then
      raise ArgumentError, "no 'from' address designation in mail"
    end

    faddr.domain || 'localhost'
  end

  def to_addrs( mail )
    toaddr = mail.destinations
    if toaddr.empty? then
      raise ArgumentError, "no 'to' address designation in mail"
    end

    toaddr
  end

  def ready_to_send( mail )
    mail.body = NKF.nkf( '-j', mail.body )
    mail.encoded
  end

end



class POP3 < Net::POP3
  include EstNet
end

class POP3client < INETclient

  RC_DEFAULT = {
    'address'  => ENV['HOSTNAME'],
    'port'     => '110',
    'account'  => ENV['USER'],
    'password' => '',
    'delete'   => true
  }


  session_type POP3

  def exec
    acc  = @rc['account']
    pass = @rc['password']
    unless acc and pass then
      CORE.warn "no account and password for popper '#{name}', cancel"
      return
    end

    do_exec acc, pass
  end

  
  private


  def if_logging_in
    @n = 0
    CORE.notice "localhost ... #{address}"
    # CORE.notice "connecting #{address}..."
  end

  def if_logged_in
    CORE.notice "localhost --- #{address}"
    # CORE.notice "connected to #{address}"

    del = @rc['delete']
    @session.each do |popmail|
      port = CORE.mkmail( CORE.mailbox.inbox ){|wp|
        popmail.mail wp
      }
      if_popmail port
      popmail.delete if del
    end
  end

  def if_popmail( port )
    port.unread = true
    port.cache = nil

    @n += 1
    s = ''
    0.upto(3) {|i| s << (@n % 3 == i ? 'o' : '-') }
    CORE.notice "localhost #{s} #{address}: #{@n}"
    # CORE.notice "pop mail \##{@n}"
  end

  def if_logged_out
    # CORE.notice "total #{@n} mails popped."
  end

end

POPclient = POP3client


class APOP < Net::APOP
  include EstNet
end

class APOPclient < POP3client

  session_type APOP

end


end   # module EST
